{# clinician_confirm_response.html #}
{% extends "base.html" %}

{% block mainsiteheader %}{% endblock %}

{% block content %}

    <h1>Thank you!</h1>

    <p>Your response has been noted. It was:</p>

    <p><b>{{ clinician_response.get_response_explanation }}</b></p>

    <p>A donation to charity will be made.</p>

    <p>You may close this web page.</p>

{% endblock %}
