#!/usr/bin/env python
# crate_anon/anonymise/print_demo_config.py

from crate_anon.anonymise.constants import DEMO_CONFIG


def main() -> None:
    print(DEMO_CONFIG)


if __name__ == '__main__':
    main()
