{# clinician_already_responded.html #}
{% comment %}
    template parameters:
        Leaflet (object)
        settings (object)
{% endcomment %}

<div class="notes">
    <h3>Background information</h3>
    <ul>
        <li>The study has been approved by the relevant Research Ethics
        Committee and the researcher has been approved and holds a
        current contract or letter of access with CPFT.</li>

        <li>A leaflet for your information:
        “<a href="{% url 'leaflet' Leaflet.CPFT_CLINRES %}"
            target="_blank"><i>Information for
        clinicians and researchers about the CPFT research
        database</i></a>”</li>

        <li>A CPFT leaflet for patients about the research database:
        “<a href="{% url 'leaflet' Leaflet.CPFT_TPIR %}"
            target="_blank"><i>Taking part in
        research</i></a>”</li>

        <li><i>Sometimes relevant:</i>
        “<a href="{% url 'leaflet' Leaflet.CPFT_TRAFFICLIGHT_CHOICE %}"
            target="_blank">Would you like to be
        contacted about research?</a>” form</li>

        <li>Donations are made to local mental health charities as a
        result of your participation by completing this form, regardless of
        your answer. Please see
        <a href="{{ settings.CHARITY_URL }}"
            target="_blank">{{ settings.CHARITY_URL_SHORT }}</a>
        for details.</li>
    </ul>

    <p>A quick reminder: <b>do not send anything patient-identifiable
    directly to the researchers.</b> Under some circumstances, this
    would give the researchers more information than the patient had
    consented to being passed on.</p>

</div>
