{# phrase_yellow.html #}

<em><b>YELLOW. I’m not sure yet. Please ask my CPFT clinicians to consider,
and perhaps pass on to me, any requests from research teams.</b>
I will choose each time whether researchers can have access to my CPFT
records and write to me.</em>
