{# query_nav.html #}

<div>
    [
    {% if nav_on_querybuilder %}
        <b>Query builder</b>
    {% else %}
        <a href="{% url 'build_query' %}">Query builder</a>
    {% endif %}
    |
    {% if nav_on_query %}
        <b>SQL</b>
    {% else %}
        <a href="{% url 'query' %}">SQL</a>
    {% endif %}
    |
    {% if nav_on_highlight %}
        <b>highlighting</b>
    {% else %}
        <a href="{% url 'highlight' %}">highlighting</a>
    {% endif %}
{% comment %}
    |
    {% if nav_on_count or not query_selected %}
        <b>preview count</b>
    {% else %}
        <a href="{% url 'count' current_query_id %}">preview count</a>
    {% endif %}
{% endcomment %}
    |
    {% if nav_on_results or not query_selected %}
        <b>results (table)</b>
    {% else %}
        <a href="{% url 'results' current_query_id %}">results (table)</a>
    {% endif %}
    |
    {% if nav_on_results_recordwise or not query_selected %}
        <b>results (recordwise)</b>
    {% else %}
        <a href="{% url 'results_recordwise' current_query_id %}">results (recordwise)</a>
    {% endif %}
    |
    {% if not query_selected %}
        <b>export as TSV</b>
    {% else %}
        <a href="{% url 'tsv' %}">export as TSV</a>
    {% endif %}
    ]
</div>
