{# page_nav.html #}

<div class="pagination">
    <span class="step-links">
        <span class="current">
            {% block current %}
            Page {{ page.number }} of {{ page.paginator.num_pages }}.
            {% endblock %}
        </span>
        [
        {% if page.has_previous %}
            <a href="?page=1">first</a>
            | <a href="?page={{ page.previous_page_number }}">previous</a>
        {% else %}
            first | previous
        {% endif %}
        |
        {% if page.has_next %}
            <a href="?page={{ page.next_page_number }}">next</a>
            | <a href="?page=-1">last</a>
        {% else %}
            next | last
        {% endif %}
        ]
    </span>
</div>
