{# base_pdf.html #}

<!DOCTYPE html> <!-- HTML 5 -->
<html>
    <head>
        <title>
            {{ RESEARCH_DB_TITLE|default:"CRATE: Research Database" }}
            letter
        </title>
        <meta charset="utf-8">
        <style type="text/css">{{ css|safe }}</style>
    </head>

    <body>
        {% block content %}{% endblock %}
    </body>

</html>
