{# query_result.html #}

{% extends "base.html" %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>Results</h1>

    {% include "expand_collapse_buttons.html" %}

    {{ table_html|safe }}

    {% include "page_nav_results.html" with page=page %}

    <h1>SQL</h1>
    <pre>{{ sql }}</pre>

{% endblock %}
