#!/usr/bin/env python
# crate_anon/crateweb/config/constants.py

CRATEWEB_CONFIG_ENV_VAR = 'CRATE_WEB_LOCAL_SETTINGS'
CHERRYPY_EXTRA_ARGS_ENV_VAR = 'CRATE_CHERRYPY_ARGS'
CELERY_APP_NAME = 'crate_anon.crateweb.consent'
