{# pid_abbrev_explanation.html #}

<div><i>
PID = patient ID (e.g. local identifier).<br>
MPID = master patient ID (e.g. NHS number).<br>
RID = research ID (hashed version of PID, consistent across database regeneration).<br>
MRID = master research ID (hashed version of MPID, consistent across database regeneration).<br>
TRID = transient research ID (integer having a 1:1 mapping to RID, that is convenient but can sometimes be regenerated).<br>
</i></div>
