{# phrase_red.html #}

<em><b>RED. I am not interested in being contacted directly by approved
researchers outside my clinical team. Please do not pass my details on to
any such researchers.</b></em>
