{# email_rdbm_new_work_pt_from_clinician.html #}
{% extends 'base_email.html' %}

{% block content %}

    <p>There’s new work on the CPFT Research Database consent-mode web
    site; a clinician wants you to send a letter on their behalf.
    Please look at the list of letters that need printing
    (it’s number <b>{{ letter.id }}</b>).
    Also, charity money needs sending.</p>

{% endblock %}
