{# page_nav_results.html #}

{% extends "page_nav.html" %}
{% load humanize %}

{% block current %}
    Page {{ page.number }} of {{ page.paginator.num_pages }}
    (<b>{{ rowcount|intcomma }}</b> rows found).
{% endblock %}
