{# database_structure.html #}

{% extends "base.html" %}

{% block content %}

    {% if paginated %}
        <h1>Research database structure: paginated</h1>
    {% else %}
        <h1>Research database structure: long form</h1>
    {% endif %}
    
    <p>Default schema: {{ default_schema }}</p>

    {% include "database_structure_table.html" %}

    {% if paginated %}
        {% include "page_nav_results.html" with page=infodictlist %}
    {% endif %}

{% endblock %}
