{# base.html #}

{% load staticfiles %}

<!DOCTYPE html> <!-- HTML 5 -->
<html>
    <head>
        <title>
            {% block title %}
                {{ RESEARCH_DB_TITLE|default:"CRATE: Research Database" }}
            {% endblock %}
        </title>
        <meta charset="utf-8">
        <link rel=StyleSheet href="{% static 'base.css' %}" type="text/css">
        <link rel=StyleSheet href="{% static 'web.css' %}" type="text/css">
        <link rel="icon" href="{% static 'favicon.ico' %}">
        {% block collapsejs %}
            <script>
                var PLUS_IMAGE = '{% static "plus.gif" %}',
                    MINUS_IMAGE = '{% static "minus.gif" %}';
            </script>
            <script type="text/javascript" src="{% static 'collapse.js' %}"></script>
        {% endblock %}
        {% block extrahead %}{% endblock %}
    </head>

    <body {% block onload %}{% endblock %}>
        {% block mainsiteheader %}
            <div>
                <i>CRATE: Clinical Records Anonymisation and Text Extraction</i>
            </div>
            <div>
                Database: <b>{{ RESEARCH_DB_TITLE }}</b>.
        
                {% block user_header %}
                    User: <b>{{ user }}</b>.
                    [
                    {% if nav_on_main_menu %}
                        Main menu
                    {% else %}
                        <a href="{% url 'home' %}">Main menu</a>
                    {% endif %}
                    | <a href="{% url 'logout' %}">Logout</a>
                    ]
                {% endblock %}
            </div>
        {% endblock %}

        {% block content %}{% endblock %}
    </body>
</html>
